@@ -55,6 +55,7 @@
 #include <osmocom/sgsn/gprs_ranap.h>
 #include <osmocom/sgsn/gprs_gmm_fsm.h>
 #include <osmocom/sgsn/gprs_mm_state_gb_fsm.h>
+#include <osmocom/sgsn/gprs_mm_state_iu_fsm.h>
 #include <osmocom/sgsn/gtp_ggsn.h>
 #include <osmocom/sgsn/gtp_mme.h>
 #include <osmocom/sgsn/sgsn_rim.h>
@@ -166,6 +167,9 @@ struct sgsn_pdp_ctx *sgsn_create_pdp_ctx(struct sgsn_ggsn_ctx *ggsn,
 	pdp->priv = pctx;
 	pctx->lib = pdp;
 
+	/* Back up our own local TEID in case we update the library one with RNC TEID when setting up Direct Tunnel: */
+	pctx->sgsn_teid_own = pdp->teid_own;
+
 	//pdp->peer =	/* sockaddr_in of GGSN (receive) */
 	//pdp->ipif =	/* not used by library */
 	pdp->version = ggsn->gtp_version;
@@ -378,6 +382,7 @@ int send_act_pdp_cont_acc(struct sgsn_pdp_ctx *pctx)
 	rc = gsm48_tx_gsm_act_pdp_acc(pctx);
 	if (rc < 0)
 		return rc;
+	pctx->ue_pdp_active = true;
 
 	if (pctx->mm->ran_type == MM_CTX_T_GERAN_Gb) {
 		/* Send SNDCP XID to MS */
@@ -464,13 +469,6 @@ reject:
 	return EOF;
 }
 
-void sgsn_pdp_upd_gtp_u(struct sgsn_pdp_ctx *pdp, void *addr, size_t alen)
-{
-	pdp->lib->gsnlu.l = alen;
-	memcpy(pdp->lib->gsnlu.v, addr, alen);
-	gtp_update_context(pdp->ggsn->gsn, pdp->lib, pdp, &pdp->lib->hisaddr0);
-}
-
 void sgsn_ggsn_echo_req(struct sgsn_ggsn_ctx *ggc)
 {
 	LOGGGSN(ggc, LOGL_INFO, "GTP Tx Echo Request\n");
@@ -567,9 +565,11 @@ static int delete_pdp_conf(struct pdp_t *pdp, void *cbp, int cause)
 			return -ENOTSUP;
 #endif
 		}
-
-		/* Confirm deactivation of PDP context to MS */
-		rc = gsm48_tx_gsm_deact_pdp_acc(pctx);
+		if (pctx->ue_pdp_active) {
+			/* Confirm deactivation of PDP context to MS */
+			rc = gsm48_tx_gsm_deact_pdp_acc(pctx);
+			pctx->ue_pdp_active = false;
+		}
 	} else {
 		LOGPDPCTXP(LOGL_NOTICE, pctx,
 			   "Not deactivating SNDCP layer since the MM context "
@@ -649,6 +649,52 @@ static int cb_conf(int type, int cause, struct pdp_t *pdp, void *cbp)
 	return 0;
 }
 
+/* Called whenever a PDP context is updated from the GGSN for any reason */
+static int cb_update_context_ind(struct pdp_t *pdp)
+{
+	struct sgsn_pdp_ctx *pctx;
+	struct sgsn_mm_ctx *mm;
+	int rc;
+
+	LOGPDPX(DGPRS, LOGL_INFO, pdp, "Context %p was updated\n", pdp);
+
+	pctx = pdp->priv;
+	if (!pctx) {
+		LOGP(DGPRS, LOGL_NOTICE,
+		     "GTP DATA IND from GGSN for unknown PDP\n");
+		return -EIO;
+	}
+	mm = pctx->mm;
+	if (!mm) {
+		LOGP(DGPRS, LOGL_ERROR,
+		     "PDP context (address=%u) without MM context!\n",
+		     pctx->address);
+		return -EIO;
+	}
+
+	if (mm->ran_type == MM_CTX_T_UTRAN_Iu) {
+#ifdef BUILD_IU
+		if (pdp->dir_tun_flags.v[0] & 0x04) { /* EI bit set ? */
+			/* GGSN informed us that it received an Error Indication when sending DL data to the RNC.
+			 * This probably means the RNC lost its state, aka crashed or was rebooted.
+			 * Page the UE so it re-creates the state at the RNC. */
+			LOGMMCTXP(LOGL_INFO, mm,
+				  "GGSN received ErrorInd from RNC while tx DL data. Paging UE in state %s\n",
+				  osmo_fsm_inst_state_name(mm->gmm_fsm));
+			rc = osmo_fsm_inst_dispatch(mm->iu.mm_state_fsm, E_PMM_RX_GGSN_GTPU_DT_EI, pctx);
+			rc = gtp_update_context_resp(sgsn->gsn, pdp,
+				 GTPCAUSE_ACC_REQ);
+			ranap_iu_page_ps(mm->imsi, &mm->p_tmsi, mm->ra.lac, mm->ra.rac);
+			return rc;
+		}
+#endif
+	}
+
+	rc = gtp_update_context_resp(sgsn->gsn, pdp,
+				 GTPCAUSE_ACC_REQ);
+	return rc;
+}
+
 /* Called whenever a PDP context is deleted for any reason */
 static int cb_delete_context(struct pdp_t *pdp)
 {
@@ -780,6 +826,8 @@ static int cb_data_ind(struct pdp_t *lib, void *packet, unsigned int len)
 #ifdef BUILD_IU
 		/* Ignore the packet for now and page the UE to get the RAB
 		 * reestablished */
+		LOGMMCTXP(LOGL_INFO, mm, "Rx GTP for UE in PMM state %s, paging it\n",
+			  osmo_fsm_inst_state_name(mm->iu.mm_state_fsm));
 		ranap_iu_page_ps(mm->imsi, &mm->p_tmsi, mm->ra.lac, mm->ra.rac);
 
 		return 0;
@@ -943,6 +991,7 @@ int sgsn_gtp_init(struct sgsn_instance *sgi)
 	}
 
 	/* Register callbackcs with libgtp */
+	gtp_set_cb_update_context_ind(gsn, cb_update_context_ind);
 	gtp_set_cb_delete_context(gsn, cb_delete_context);
 	gtp_set_cb_conf(gsn, cb_conf);
 	gtp_set_cb_recovery3(gsn, cb_recovery3);
